
/* fancy ihex .c */

/*
 * This file: fihex.c
 *
 * Date:
 *
 * Reads and splits intel hex files.
 *
 * Software from Tenison Technology
 * (C) 1998 Tenison Technology 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define uchar unsigned char
#define baseaddr 0x4000

const sizer = 1234567;
FILE *fd;
int oddevf = 0;
int pokef = 0;
char *opfname = NULL;

int iread(len)
{
  int r = 0;
  int i;
  for (i=0; i < len; i++)
  {
    unsigned char c = getc(fd);
    c = (c <= '9') ? c-'0': c-('0'+7);
    r = (r<<4) + c;
  }
  return r;
}


static char *pp, *base, *limit;

read_in_data(uchar c)
{
  *pp ++ = c;
}


verify(uchar c)
{
  uchar d = *pp;
  if (d != c)
    {
      printf("Verify error at 0x%x   %02x  %02x\n", pp-base, c, d);
    }
  pp ++;
  if (pp > limit) 
    {
      printf("Second file is longer");
      exit(1);
    }
}



void pokedump(char *buffer, int sa, int len)
{
  int b = 0;
  int addr = sa;
  int mastersum = 0;
  FILE *fd;
  if (opfname) fd = fopen(opfname, "w"); else fd = stdout;
  if (fd == NULL)
    {
      perror("");
      exit(1);
    }
  fprintf(fd, "M%04x     ",  baseaddr);
  while (len > 0)
    {
      uchar d = buffer[b++];
      uchar e = buffer[b++];
      fprintf(fd, "%04x", (e<<8)+d);
      mastersum += d + e;
      len -= 2;
    }
  fprintf(fd,".\n");
  fprintf(stderr, "Master sum %08X\n", mastersum);
}


void ihdump(char *buffer, int sa, int len)
{
  int b = 0;
  int addr = sa;
  int mastersum = 0;
  FILE *fd;
  if (opfname) fd = fopen(opfname, "w"); else fd = stdout;
  if (fd == NULL)
    {
      perror("");
      exit(1);
    }
  while (len > 0)
    {
      fprintf(fd, "%c\n", 13);
      {
	uchar sum;
	int i;
	int objbytes = 24;
	fprintf(fd, ":%02X%04X00", objbytes, addr);
	sum = objbytes + (addr & 255) + ((addr >> 8) & 255);
	for (i=0; i<objbytes; i++) 
	  {
	    int a = b++;
	    uchar d;
	    if (oddevf & 2) a = a+a;
	    if (oddevf & 1) a = a+a+1;

	    d = buffer[a];
	    addr ++;
	    fprintf(fd, "%02X", d);
	    sum = d + sum;
	    mastersum += d;
	    if (oddevf) len -= 2; else len -= 1;
	  }
	fprintf(fd, "%02X%c\n", (256 - sum) & 255, 13);
      }
    }
  fprintf(stderr, "Master sum %08X\n", mastersum);
}

int main (argc, argv) char ** argv;
{
  char * buffer = (char *) malloc(sizer);
  if (argc == 1)
    {
      printf("usenix:\n %s [-pokef -odd -even] [-o fn]  fn\n", argv[0]);
      exit(1);
    }
  while(1)
    {
      if (!strcmp(argv[1], "-odd"))
	{
	  oddevf = 1;
	  argc --; argv ++;
	  continue;
	}
     
      if (!strcmp(argv[1], "-poke"))
	{
	  pokef = 1;
	  argc --; argv ++;
	  continue;
	}
      if (!strcmp(argv[1], "-even"))
	{
	  oddevf = 2;
	  argc --; argv ++;
	  continue;
	}
      if (!strcmp(argv[1], "-o"))
	{
	  opfname = strdup(argv[2]);
	  argc -=2; argv +=2;
	  continue;
	}
      break;
    }


  pp = buffer;
  base = buffer;
  limit = buffer+sizer;
  run(argv[1], read_in_data);

  if (pokef) pokedump(buffer, 0, pp-buffer);
  else ihdump(buffer, 0, pp-buffer);

  return 0;

}


int bigsum(unsigned char *l, int b)
{
  int r = 0;
  while (b-- > 0)
    {
      r += *l++;
    }
  return r;
}


run(fn, f)
char *fn;
void (*f)();
{
  int line = 1;
  int addr = -1;
  int bytes = 0;  
  fd = fopen(fn, "r");
  if (fd == 0) 
    {
      perror("cant open input file");
      exit(1);
    }

  while (feof(fd)==0)
  {
    uchar c = getc(fd);
    int llen, laddr, ltype;
    uchar lsum;
    int wq;

    if (c == '\n') line++;
    if (c != ':') continue;
    llen = iread(2);
    laddr = iread(4);
    ltype = iread(2);
/*
    printf("Type %x  addr %x len %x \n", ltype, laddr, llen);
*/
    if (ltype != 0) 
    { 
      printf("Type %i encountered in line %i\n", ltype, line);
      continue;
    }
    if (addr == -1) addr = laddr;
    else if (laddr != addr)
    {
      printf("Disjoint addresses %x and %x in line %i\n",
                   laddr, addr, line);
      addr = laddr;
    }
    lsum = llen + laddr + ltype + (laddr >> 8);
    for (wq=0; wq<llen; wq++)
    {
      uchar data = iread(2);
      lsum = lsum + data;
      f(data);
      bytes += 1;
    }
    lsum = lsum + iread(2);

    if (lsum != 0) 
          fprintf(stderr, "Checksum error in line %i, residue %i\n", line, lsum);    

    addr = addr + llen;
  }
  /*  printf("Bytes read = %i bytes.\n", bytes);
   */
  return 0;
}

